%{

/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


extern int yydebug ;

#define YYDEBUG 1
#define EXTERN
#define EXTERN_INIT(x,y) x = y

#include <errno.h>
#include "basename.hxx"
#include "yoocom.hxx"

#define SIZEOF_MESS 256 

static int SvcErrorFlag ;
static char SvcError[SIZEOF_MESS] ;

char* infile = NULL ;
char* progname = NULL ;

extern "C" {
  int yywrap() ;
  int yyparse() ;
}

#if defined(__alpha__) || (defined(__sun__) && !defined(__svr4__))
int yylex() ;
#endif

#if defined(__alpha__) || (defined(__sun__) && !defined(__svr4__))
int yyoutput(int c) ; 
#else
extern "C" { int yyoutput(int c) ; }
#endif

// CONTOURNEMENT D'UN BUG DU COMPILO G++ 2.5.8
static ListPtr<SvcMember> dummy_member ;
static IterListPtr<SvcMember> dummy_member_iter(dummy_member) ;
static ListPtr<SvcIdent> dummy_ident ;
static IterListPtr<SvcIdent> dummy_ident_iter(dummy_ident) ;


%}

%union {
  char                ident[SIZEOF_TOKEN] ;
  int                 value ;
  SvcClass*           classdecl ;
  ListPtr<SvcIdent>*  identlist ;
  SvcIdent*           attrident ;
  ListPtr<SvcClass>*  classlist ;
  SvcHead*            classhead ;
  ListPtr<SvcMember>* memberlist ;
  SvcMember*          member ;
  SvcType*            typespec ;
  ListPtr<SvcParam>*  paramlist ;
  SvcParam*           param ;
} ;


%token <ident> IDENTIFIER
%token <ident> CHAR U_CHAR SHORT U_SHORT INT U_INT FLOAT DOUBLE STRING VOID
%token <value> IN 0 OUT 1 VINT CONST
%token REMOTE ACCESS SVCSTRUCT SVCCLASS FLOW TABLE CBRACKET OBRACKET VIRTUAL
%token EQ_0

%type  <classdecl> base_list base_class
%type  <classhead> svcclass_head
%type  <memberlist> member_list remote_member_list local_member_list
%type  <memberlist> multiple_local_member properties
%type  <member> remote_member_function
%type  <typespec> complex_spec flow_spec table_spec template_spec svctype_spec 
%type  <paramlist> parameter_list
%type  <param> parameter
%type  <identlist> ident_list
%type  <ident> svcclass_name svcstruct_head
%type  <value> inout_spec
%type  <attrident> identifier

%start all

%%

all :
        svcschema 

	
svcschema :
        /* empty */
        | svcclass CBRACKET svcschema
        | svcstruct CBRACKET svcschema

	
svcclass : 
        svcclass_head  member_list 
          { 
	    SvcClass* cl = new SvcClass($1, $2) ;             
            cl->cppGenServerRemote() ;
            yyoutput('}') ;
	    SchemaClass.insert(cl, -1) ; 
	  } 
	| svcclass_head local_member_list member_list
          { 
	    IterListPtr<SvcMember> iter(*$2) ;
	    for (; iter.more(); iter.next()) {
	      $3->insert(iter.get(), -1) ;
	    }

	    SvcClass* cl = new SvcClass($1, $3) ;
            cl->cppGenServerRemote() ;
            yyoutput('}') ;
	    SchemaClass.insert(cl, -1) ; 
	  } ;


svcstruct : 
	svcstruct_head local_member_list
          { 
	    SvcStruct* cl = new SvcStruct($1, $2) ;
            yyoutput('}') ;
	    SchemaStruct.insert(cl, -1) ; 
	  } ;


svcclass_head : 
	SVCCLASS svcclass_name {yyoutput(':'); cppGenServerHeader();} OBRACKET
          { 
            yyoutput('{') ;
	    $$ = new SvcHead($2) ;
            $$->cppGenCtor("SvcServer"); 
	  }
	| SVCCLASS svcclass_name ':' base_list OBRACKET
          { 
	    if ($4 == NULL) { 
	      yyoutput(',') ;
	      cppGenServerHeader(); 
	    }
            yyoutput('{') ;
            $$ = new SvcHead($2, $4) ;
            char* b = ($4 == NULL) ? "SvcServer" : $4->classname();
            $$->cppGenCtor(b); 
          } ;  


svcstruct_head :
          SVCSTRUCT svcclass_name OBRACKET
          { 
            yyoutput('{') ;
            strcpy($$,$2) ;
	  } ;


svcclass_name :
	IDENTIFIER { strcpy($$, $1) ; } ;


base_list :
        base_class
          {
	    $$ = $1 ;
	  }
        | base_class ',' base_list
          {
	    $$ = $1 != NULL ? $1 : $3 ;
	  }


base_class : 
        inherit_mode svcclass_name
          {
	    $$ = findClass($2) ;
	  }


inherit_mode :
        /* empty */
        | ACCESS 
        | ACCESS VIRTUAL 
        | VIRTUAL ACCESS 
        | VIRTUAL ;

	
member_list :
        /* empty */
          {
	    $$ = new ListPtr<SvcMember> ;
	  }
	| REMOTE ':' remote_member_list member_list
          {
	    IterListPtr<SvcMember> iter(*$3) ;
	    for (; iter.more(); iter.next()) {
	      $4->insert(iter.get(), 0) ;
	    }
	    $$ = $4 ;
	  }
	| ACCESS ':' local_member_list member_list
          {
	    IterListPtr<SvcMember> iter(*$3) ;
	    for (; iter.more(); iter.next()) {
	      $4->insert(iter.get(), 0) ;
	    }
	    $$ = $4 ;
	  } ;


remote_member_list :
        /* empty */
          {
	    $$ = new ListPtr<SvcMember> ;
	  }
	| remote_member_function remote_member_list
          {
	    $2->insert($1, 0) ;
	    $$ = $2 ;
	  } ;


remote_member_function :
	vtl_kwd complex_spec IDENTIFIER '(' ')' ';'
          {
	    $$ = new SvcRemoteMember($2, $3, NULL) ;
	  } ;
	| vtl_kwd complex_spec IDENTIFIER '(' ')' EQ_0';'
          {
	    $$ = new SvcRemoteMember($2, $3, NULL) ;
	  } ;
	| vtl_kwd complex_spec IDENTIFIER '(' parameter_list ')' ';'
          {
	    $$ = new SvcRemoteMember($2, $3, $5) ;
	  } 
	| vtl_kwd complex_spec IDENTIFIER '(' parameter_list ')' EQ_0 ';'
          {
	    $$ = new SvcRemoteMember($2, $3, $5) ;
	  } 
        | error ';' { YYABORT ; } ;

vtl_kwd :
        /* empty */
        | VIRTUAL ;

complex_spec :
	svctype_spec    { $$ = new SvcSimple($1) ; } 
	| flow_spec     { $$ = $1 ; } 
	| table_spec     { $$ = $1 ; } 
	| template_spec { $$ = $1 ; } ;


flow_spec :
	FLOW '<' svctype_spec '>'
          {
	    $$ = new SvcFlow($3) ;
	  } 


table_spec :
	TABLE '<' svctype_spec '>'
          {
	    $$ = new SvcTable($3) ;
	  } 


template_spec :
	IDENTIFIER '<' svctype_spec '>'
          {
	    $$ = new SvcTemplate($1, $3) ;
	  } 


svctype_spec :
	IDENTIFIER  { $$ = new SvcType($1, T_OTHER) ; } 
	| CHAR     { $$ = new SvcType($1, T_CHAR) ; } 
	| U_CHAR   { $$ = new SvcType($1, T_UCHAR) ; } 
	| SHORT    { $$ = new SvcType($1, T_SHORT) ; } 
	| U_SHORT  { $$ = new SvcType($1, T_USHORT) ; } 
	| INT      { $$ = new SvcType($1, T_INT) ; } 
	| U_INT    { $$ = new SvcType($1, T_UINT) ; } 
	| FLOAT    { $$ = new SvcType($1, T_FLOAT) ; } 
	| DOUBLE   { $$ = new SvcType($1, T_DOUBLE) ; } 
	| STRING   { $$ = new SvcType($1, T_STRING) ; } 
        | VOID     { $$ = new SvcType($1, T_VOID) ; } ;
	

parameter_list :
	parameter
          {
	    $$ = new ListPtr<SvcParam> ;
	    $$->insert($1, 0) ; 
	  }
	| parameter ',' parameter_list
          {
	    $3->insert($1, 0) ;
	    $$ = $3 ;
	  } ;


parameter :
	inout_spec complex_spec '&' IDENTIFIER
          {
	    $$ = new SvcParam($1, $2, $4) ;
	  } 
	| inout_spec complex_spec IDENTIFIER
          {
	    $$ = new SvcParam($1, $2, $3) ;
	  } ;


inout_spec :
	/* empty */ { $$ = IN ; }
	| CONST     { $$ = IN ; }
	| IN        { $$ = $1 ; }
	| IN CONST  { $$ = $1 ; }  
	| OUT       { $$ = $1 ; } ;


local_member_list :
        multiple_local_member
          {
	    $$ = new ListPtr<SvcMember> ;
	    if ($1 != NULL) {
	      IterListPtr<SvcMember> iter(*$1) ;
	      for (; iter.more(); iter.next()) {
		$$->insert(iter.get(), 0) ;
	      }
	    }	    
	  }
	| multiple_local_member local_member_list
          {
	    if ($1 != NULL) {
	      IterListPtr<SvcMember> iter(*$1) ;
	      for (; iter.more(); iter.next()) {
		$2->insert(iter.get(), 0) ;
	      }
	    }
	    $$ = $2 ;
	  }


multiple_local_member :
	properties { $$ = $1 ; }
	| error ';' 
          {
            SvcErrorFlag = 0 ;
	    yyerrok ;
	    $$ = NULL;
	  } ;


properties :
	complex_spec ident_list ';' 
          { 
	    $$ = new ListPtr<SvcMember> ;
	    IterListPtr<SvcIdent> iter(*$2) ;
	    for (; iter.more(); iter.next()) {
	      if (iter.get()->size() == 0) {
		$$->insert(new SvcLocalMember($1, iter.get()->name()), 0) ;
	      } else {
		$$->insert(new SvcMultipleLocalMember($1,iter.get()->name(),
						      iter.get()->size()), 0);
	      }
	    }  
	  } ;


ident_list :
	identifier
          {
	    $$ = new ListPtr<SvcIdent> ;
	    $$->insert($1, 0) ;
	  }
	| identifier ',' ident_list
          {
	    $3->insert($1, 0) ;
	    $$ = $3;
	  } ;

identifier :
        IDENTIFIER
          {
     	    $$ = new SvcIdent($1, 0) ;
          }
        | IDENTIFIER '[' VINT ']'
          {
     	    $$ = new SvcIdent($1, $3) ;
          }


%%

#ifndef hpux
extern "C" {
#endif

void yyerror(char* s) 
{
  if (SvcErrorFlag) {    
    printf("%s", SvcError) ;
    exit(1) ;
  } else {
    sprintf(SvcError,"%s:%d: %s\n", infile, linecount+1,s) ;
    SvcErrorFlag = 1 ;
  }
}

#ifndef hpux
}
#endif

void option_error(char* name) 
{
  printf("usage: %s [-v] -c++|-space -client|-server [-connected|-notconnected] <input> -o <output>\n",
	 name);
  exit(1) ;
}


void headerfile(FILE* f, char* name)
{
  fprintf(yyout, "#ifndef __sockrpc_HXX\n") ;
  fprintf(yyout, "#include \"sockrpc.hxx\"\n") ;
  fprintf(yyout, "#endif\n\n") ;
  char* base = yoo_basename(name) ;
  fprintf(yyout, "#ifndef __%s_HXX\n", base) ;
  fprintf(yyout, "#define __%s_HXX\n\n", base) ;
}


int main(int argc, char** argv) {
  
  progname = argv[0] ;
  char* outfile = NULL ;
  int vers = 0 ;
  
  for (int i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-connected") == 0) {
      if (conn_gen != -1) option_error(argv[0]) ;
      conn_gen = CONNECTED ;
    } else if (strcmp(argv[i], "-notconnected") == 0) {
      if (conn_gen != -1) option_error(argv[0]) ;
      conn_gen = DISCONNECTED ;
    } else if (strcmp(argv[i], "-c++") == 0) {
      if (lang_gen != -1) option_error(argv[0]) ;
      lang_gen = CPP ;
    } else if (strcmp(argv[i], "-space") == 0) {
      if (lang_gen != -1) option_error(argv[0]) ;
      lang_gen = ASL ;
    } else if (strcmp(argv[i], "-client") == 0) {
      if (side_gen != -1) option_error(argv[0]) ;
      side_gen = CLIENT ;
    } else if (strcmp(argv[i], "-server") == 0) {
      if (side_gen != -1) option_error(argv[0]) ;
      side_gen = SERVER ;
    } else if (strcmp(argv[i], "-o") == 0) {
      if (outfile != NULL) option_error(argv[0]) ;
      i++ ;
      if (i < argc) outfile = argv[i] ;
    } else if (strcmp(argv[i], "-v") == 0) {
      vers = 1 ;
    } else {
      if (infile != NULL) option_error(argv[0]) ;
      infile = argv[i] ;
    }
  }

  if (conn_gen == -1) conn_gen = CONNECTED ;
  
  if ( (lang_gen == -1) || (side_gen == -1) || (outfile == NULL) ||
       (infile == NULL) || (conn_gen == -1) ) {
    option_error(argv[0]) ;
  }

  if (vers) {
    printf("YOODA version 1.2 <%s>, Copyright (C) 1994 Eric Abecassis
YOODA comes with ABSOLUTELY NO WARRANTY; for details type `COPYING'
in YOODA_HOME directory. This is free software, and you are welcome to 
redistribute it under certain conditions; for details read `COPYING'
or 'COPYING.LIB' in YOODA_HOME directory\n", argv[0]) ;
  }

  yyin = fopen(infile, "r") ;
  if ((side_gen == SERVER) && (lang_gen == CPP)) {
    yyout = fopen(outfile, "w") ;
    headerfile(yyout, outfile) ;
  } else {
    yyout = fopen("/dev/null", "rw") ;
  }
  if ( (!yyin) || (!yyout) ) {
    perror("") ;
    exit(1) ;
  }
  
  linecount = 0 ;
  SvcErrorFlag = 0 ;
  yyparse() ;
  yyerror("") ;
  if ((side_gen == SERVER) && (lang_gen == CPP)) {
    fprintf(yyout, "\n\n#endif\n") ;
    fclose(yyout) ;
  } else {
    yyout = fopen(outfile, "w") ;  
    if ( (lang_gen == CPP) && (side_gen == CLIENT) ) {
      headerfile(yyout, outfile) ;
      cppGenClient() ;
      fprintf(yyout, "\n\n#endif\n") ;
    } else if ( (lang_gen == ASL) && (side_gen == CLIENT) ) {
      aslGenClient() ;
    } else if ( (lang_gen == ASL) && (side_gen == SERVER) ) {
      aslGenServer() ;
    }
    fclose(yyout) ;
  }
}
	 
